# -*- coding: utf-8 -*-

"""
***************************************************************************
    QtCore.py
    ---------------------
    Date                 : November 2015
    Copyright            : (C) 2015 by Matthias Kuhn
    Email                : matthias at opengis dot ch
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = 'Matthias Kuhn'
__date__ = 'November 2015'
__copyright__ = '(C) 2015, Matthias Kuhn'

from PyQt6.QtCore import *

from types import MethodType


_QVariant__repr__ = QVariant.__repr__
_QVariant__eq__ = QVariant.__eq__
_QVariant__ne__ = QVariant.__ne__
_QVariant__hash__ = QVariant.__hash__


def __bool__(self):
    return not self.isNull()


def __repr__(self):
    if self.isNull():
        return 'NULL'
    else:
        return _QVariant__repr__(self)


def __eq__(self, other):
    if self.isNull():
        return (isinstance(other, QVariant) and other.isNull()) or other is None
    else:
        return _QVariant__eq__(self, other)


def __ne__(self, other):
    if self.isNull():
        return not (isinstance(other, QVariant) and other.isNull()) and other is not None
    else:
        return _QVariant__ne__(self, other)


def __hash__(self):
    if self.isNull():
        return 2178309
    else:
        return _QVariant__hash__(self)


QVariant.__bool__ = __bool__
QVariant.__repr__ = __repr__
QVariant.__eq__ = __eq__
QVariant.__ne__ = __ne__
QVariant.__hash__ = __hash__

if (QT_VERSION >= 0x060000):
    NULL = None
    # monkey patch in QVariant Types
    QVariant.Invalid = QMetaType.Type.UnknownType
    QVariant.Bool = QMetaType.Type.Bool
    QVariant.Int = QMetaType.Type.Int
    QVariant.UInt = QMetaType.Type.UInt
    QVariant.LongLong = QMetaType.Type.LongLong
    QVariant.ULongLong = QMetaType.Type.ULongLong
    QVariant.Double = QMetaType.Type.Double
    QVariant.Char = QMetaType.Type.QChar
    QVariant.Map = QMetaType.Type.QVariantMap
    QVariant.List = QMetaType.Type.QVariantList
    QVariant.String = QMetaType.Type.QString
    QVariant.StringList = QMetaType.Type.QStringList
    QVariant.ByteArray = QMetaType.Type.QByteArray
    QVariant.BitArray = QMetaType.Type.QBitArray
    QVariant.Date = QMetaType.Type.QDate
    QVariant.Time = QMetaType.Type.QTime
    QVariant.DateTime = QMetaType.Type.QDateTime
    QVariant.Url = QMetaType.Type.QUrl
    QVariant.Locale = QMetaType.Type.QLocale
    QVariant.Rect = QMetaType.Type.QRect
    QVariant.RectF = QMetaType.Type.QRectF
    QVariant.Size = QMetaType.Type.QSize
    QVariant.SizeF = QMetaType.Type.QSizeF
    QVariant.Line = QMetaType.Type.QLine
    QVariant.LineF = QMetaType.Type.QLineF
    QVariant.Point = QMetaType.Type.QPoint
    QVariant.PointF = QMetaType.Type.QPointF
    QVariant.RegularExpression = QMetaType.Type.QRegularExpression
    QVariant.Hash = QMetaType.Type.QVariantHash
    QVariant.EasingCurve = QMetaType.Type.QEasingCurve
    QVariant.Uuid = QMetaType.Type.QUuid
    QVariant.ModelIndex = QMetaType.Type.QModelIndex
    QVariant.PersistentModelIndex = QMetaType.Type.QPersistentModelIndex
    QVariant.LastCoreType = QMetaType.Type.LastCoreType
    QVariant.Font = QMetaType.Type.QFont
    QVariant.Pixmap = QMetaType.Type.QPixmap
    QVariant.Brush = QMetaType.Type.QBrush
    QVariant.Color = QMetaType.Type.QColor
    QVariant.Palette = QMetaType.Type.QPalette
    QVariant.Image = QMetaType.Type.QImage
    QVariant.Polygon = QMetaType.Type.QPolygon
    QVariant.Region = QMetaType.Type.QRegion
    QVariant.Bitmap = QMetaType.Type.QBitmap
    QVariant.Cursor = QMetaType.Type.QCursor
    QVariant.KeySequence = QMetaType.Type.QKeySequence
    QVariant.Pen = QMetaType.Type.QPen
    QVariant.TextLength = QMetaType.Type.QTextLength
    QVariant.TextFormat = QMetaType.Type.QTextFormat
    QVariant.Transform = QMetaType.Type.QTransform
    QVariant.Matrix4x4 = QMetaType.Type.QMatrix4x4
    QVariant.Vector2D = QMetaType.Type.QVector2D
    QVariant.Vector3D = QMetaType.Type.QVector3D
    QVariant.Vector4D = QMetaType.Type.QVector4D
    QVariant.Quaternion = QMetaType.Type.QQuaternion
    QVariant.PolygonF = QMetaType.Type.QPolygonF
    QVariant.Icon = QMetaType.Type.QIcon
    QVariant.SizePolicy = QMetaType.Type.QSizePolicy
    QVariant.UserType = QMetaType.Type.User

    from enum import Enum


    def _force_int(v): return int(v.value) if isinstance(v, Enum) else v

    QMetaType.Type.__int__ = _force_int
    QMetaType.Type.__eq__ = lambda t1, t2: _force_int(t1) == _force_int(t2)

    # These types aren't IntEnums or IntFlags, so patch that back in
    # See discussion at https://www.riverbankcomputing.com/pipermail/pyqt/2024-February/045715.html
    Qt.CheckState.__int__ = _force_int
    Qt.CheckState.__eq__ = lambda t1, t2: _force_int(t1) == _force_int(t2)

    # patch back in Qt flags removed in PyQt

    QAbstractItemModel.CheckIndexOptions = lambda flags=0: QAbstractItemModel.CheckIndexOption(flags)
    QByteArray.Base64Options = lambda flags=0: QByteArray.Base64Option(flags)
    QCommandLineOption.Flags = lambda flags=0: QCommandLineOption.Flag(flags)
    QDir.Filters = lambda flags=0: QDir.Filter(flags)
    QDir.SortFlags = lambda flags=0: QDir.SortFlag(flags)
    QDirIterator.IteratorFlags = lambda flags=0: QDirIterator.IteratorFlag(flags)
    QEventLoop.ProcessEventsFlags = lambda flags=0: QEventLoop.ProcessEventsFlag(flags)
    QFileDevice.FileHandleFlags = lambda flags=0: QFileDevice.FileHandleFlag(flags)
    QFileDevice.MemoryMapFlags = lambda flags=0: QFileDevice.MemoryMapFlag(flags)
    QFileDevice.Permissions = lambda flags=0: QFileDevice.Permission(flags)
    QItemSelectionModel.SelectionFlags = lambda flags=0: QItemSelectionModel.SelectionFlag(flags)
    QLibrary.LoadHints = lambda flags=0: QLibrary.LoadHint(flags)
    QLocale.DataSizeFormats = lambda flags=0: QLocale.DataSizeFormat(flags)
    QLocale.NumberOptions = lambda flags=0: QLocale.NumberOption(flags)
    QMetaType.TypeFlags = lambda flags=0: QMetaType.TypeFlag(flags)
    QRegularExpression.MatchOptions = lambda flags=0: QRegularExpression.MatchOption(flags)
    QRegularExpression.PatternOptions = lambda flags=0: QRegularExpression.PatternOption(flags)
    QStandardPaths.LocateOptions = lambda flags=0: QStandardPaths.LocateOption(flags)
    QTextBoundaryFinder.BoundaryReasons = lambda flags=0: QTextBoundaryFinder.BoundaryReason(flags)
    QTextStream.NumberFlags = lambda flags=0: QTextStream.NumberFlag(flags)
    QUrl.ComponentFormattingOptions = lambda flags=0: QUrl.ComponentFormattingOption(flags)
    QUrl.FormattingOptions = lambda flags=0: QUrl.UrlFormattingOption(flags)
    QUrl.UserInputResolutionOptions = lambda flags=0: QUrl.UserInputResolutionOption(flags)
    Qt.Alignment = lambda flags=0: Qt.AlignmentFlag(flags)
    Qt.ApplicationStates = lambda flags=0: Qt.ApplicationState(flags)
    Qt.DockWidgetAreas = lambda flags=0: Qt.DockWidgetArea(flags)
    Qt.DropActions = lambda flags=0: Qt.DropAction(flags)
    Qt.Edges = lambda flags=0: Qt.Edge(flags)
    Qt.FindChildOptions = lambda flags=0: Qt.FindChildOption(flags)
    Qt.GestureFlags = lambda flags=0: Qt.GestureFlag(flags)
    Qt.ImageConversionFlags = lambda flags=0: Qt.ImageConversionFlag(flags)
    Qt.InputMethodHints = lambda flags=0: Qt.InputMethodHint(flags)
    Qt.InputMethodQueries = lambda flags=0: Qt.InputMethodQuery(flags)
    Qt.ItemFlags = lambda flags=0: Qt.ItemFlag(flags)
    Qt.KeyboardModifiers = lambda flags=0: Qt.KeyboardModifier(flags)
    Qt.MatchFlags = lambda flags=0: Qt.MatchFlag(flags)
    Qt.MouseButtons = lambda flags=0: Qt.MouseButton(flags)
    Qt.MouseEventFlags = lambda flags=0: Qt.MouseEventFlag(flags)
    Qt.Orientations = lambda flags=0: Qt.Orientation(flags)
    Qt.ScreenOrientations = lambda flags=0: Qt.ScreenOrientation(flags)
    Qt.TextInteractionFlags = lambda flags=0: Qt.TextInteractionFlag(flags)
    Qt.ToolBarAreas = lambda flags=0: Qt.ToolBarArea(flags)
    Qt.TouchPointStates = lambda flags=0: Qt.TouchPointState(flags)
    Qt.WindowStates = lambda flags=0: Qt.WindowState(flags)
    Qt.WindowFlags = lambda flags=0: Qt.WindowType(flags)
else:
    NULL = QVariant(QVariant.Int)



"""
***************************************************************************
    ExecuteSQL.py -- use virtual layers to execute SQL on any sources
    ---------------------
    Date                 : Jan 2016
    Copyright            : (C) 2016 by Hugo Mercier
    Email                : hugo dot mercier at oslandia dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Hugo Mercier"
__date__ = "January 2016"
__copyright__ = "(C) 2016, Hugo Mercier"

from qgis.core import (
    Qgis,
    QgsExpression,
    QgsFeatureSink,
    QgsProcessingAlgorithm,
    QgsProcessingException,
    QgsProcessingParameterCrs,
    QgsProcessingParameterDefinition,
    QgsProcessingParameterEnum,
    QgsProcessingParameterFeatureSink,
    QgsProcessingParameterMultipleLayers,
    QgsProcessingParameterString,
    QgsProcessingUtils,
    QgsProject,
    QgsVectorFileWriter,
    QgsVectorLayer,
    QgsVirtualLayerDefinition,
    QgsWkbTypes,
)

from processing.algs.qgis.QgisAlgorithm import QgisAlgorithm


class ParameterExecuteSql(QgsProcessingParameterDefinition):
    def __init__(self, name="", description=""):
        super().__init__(name, description)
        self.setMetadata(
            {
                "widget_wrapper": "processing.algs.qgis.ui.ExecuteSQLWidget.ExecuteSQLWidgetWrapper"
            }
        )

    def type(self):
        return "execute_sql"

    def clone(self):
        return ParameterExecuteSql(self.name(), self.description())


class ExecuteSQL(QgisAlgorithm):
    """This algorithm allows executing an SQL query on a set of input
    vector layers thanks to the virtual layer provider
    """

    INPUT_DATASOURCES = "INPUT_DATASOURCES"
    INPUT_QUERY = "INPUT_QUERY"
    INPUT_UID_FIELD = "INPUT_UID_FIELD"
    INPUT_GEOMETRY_FIELD = "INPUT_GEOMETRY_FIELD"
    INPUT_GEOMETRY_TYPE = "INPUT_GEOMETRY_TYPE"
    INPUT_GEOMETRY_CRS = "INPUT_GEOMETRY_CRS"
    OUTPUT = "OUTPUT"

    def group(self):
        return self.tr("Vector general")

    def groupId(self):
        return "vectorgeneral"

    def __init__(self):
        super().__init__()

        self.geometry_types = [
            (None, self.tr("Autodetect")),
            (Qgis.WkbType.NoGeometry, self.tr("No geometry")),
            (Qgis.WkbType.Point, self.tr("Point")),
            (Qgis.WkbType.LineString, self.tr("LineString")),
            (Qgis.WkbType.Polygon, self.tr("Polygon")),
            (Qgis.WkbType.MultiPoint, self.tr("MultiPoint")),
            (Qgis.WkbType.MultiLineString, self.tr("MultiLineString")),
            (Qgis.WkbType.MultiPolygon, self.tr("MultiPolygon")),
        ]

    def flags(self):
        return super().flags() | QgsProcessingAlgorithm.Flag.FlagNoThreading

    def initAlgorithm(self, config=None):
        self.addParameter(
            QgsProcessingParameterMultipleLayers(
                name=self.INPUT_DATASOURCES,
                description=self.tr(
                    "Input data sources (called input1, .., inputN in the query)"
                ),
                layerType=Qgis.ProcessingSourceType.Vector,
                optional=True,
            )
        )

        self.addParameter(
            ParameterExecuteSql(name=self.INPUT_QUERY, description=self.tr("SQL query"))
        )

        self.addParameter(
            QgsProcessingParameterString(
                name=self.INPUT_UID_FIELD,
                description=self.tr("Unique identifier field"),
                optional=True,
            )
        )

        self.addParameter(
            QgsProcessingParameterString(
                name=self.INPUT_GEOMETRY_FIELD,
                description=self.tr("Geometry field"),
                optional=True,
            )
        )

        self.addParameter(
            QgsProcessingParameterEnum(
                self.INPUT_GEOMETRY_TYPE,
                self.tr("Geometry type"),
                options=[t[1] for t in self.geometry_types],
                defaultValue=0,
            )
        )

        self.addParameter(
            QgsProcessingParameterCrs(
                self.INPUT_GEOMETRY_CRS, self.tr("CRS"), optional=True
            )
        )

        self.addParameter(
            QgsProcessingParameterFeatureSink(self.OUTPUT, self.tr("SQL Output"))
        )

    def name(self):
        return "executesql"

    def displayName(self):
        return self.tr("Execute SQL")

    def shortDescription(self):
        return self.tr("Runs a query with SQL syntax.")

    def shortHelpString(self):
        return self.tr(
            "This algorithm runs a query with SQL syntax.\n"
            "Input data sources are identified with input1, input2, ..., inputN "
            "and a simple query will look like: 'SELECT * FROM input1'.\n"
            "The result of the query will be added as a new layer."
        )

    def processAlgorithm(self, parameters, context, feedback):
        layers = self.parameterAsLayerList(parameters, self.INPUT_DATASOURCES, context)
        query = self.parameterAsString(parameters, self.INPUT_QUERY, context)
        uid_field = self.parameterAsString(parameters, self.INPUT_UID_FIELD, context)
        geometry_field = self.parameterAsString(
            parameters, self.INPUT_GEOMETRY_FIELD, context
        )
        geometry_type = self.geometry_types[
            self.parameterAsEnum(parameters, self.INPUT_GEOMETRY_TYPE, context)
        ][0]
        geometry_crs = self.parameterAsCrs(parameters, self.INPUT_GEOMETRY_CRS, context)

        df = QgsVirtualLayerDefinition()
        for layerIdx, layer in enumerate(layers):
            # Issue https://github.com/qgis/QGIS/issues/24041
            # When using this algorithm from the graphic modeler, it may try to
            # access (thanks the QgsVirtualLayerProvider) to memory layer that
            # belongs to temporary QgsMapLayerStore, not project.
            # So, we write them to disk is this is the case.
            if context.project() and not context.project().mapLayer(layer.id()):
                basename = (
                    "memorylayer." + QgsVectorFileWriter.supportedFormatExtensions()[0]
                )
                tmp_path = QgsProcessingUtils.generateTempFilename(basename, context)
                QgsVectorFileWriter.writeAsVectorFormat(
                    layer, tmp_path, layer.dataProvider().encoding()
                )
                df.addSource(f"input{layerIdx + 1}", tmp_path, "ogr")
            else:
                df.addSource(f"input{layerIdx + 1}", layer.id())

        if query == "":
            raise QgsProcessingException(
                self.tr("Empty SQL. Please enter valid SQL expression and try again.")
            )
        localContext = self.createExpressionContext(parameters, context)
        expandedQuery = QgsExpression.replaceExpressionText(query, localContext)
        df.setQuery(expandedQuery)

        if uid_field:
            df.setUid(uid_field)

        if geometry_type == Qgis.WkbType.NoGeometry:
            df.setGeometryWkbType(Qgis.WkbType.NoGeometry)
        else:
            if geometry_field:
                df.setGeometryField(geometry_field)
            if geometry_type is not None:
                df.setGeometryWkbType(geometry_type)
            if geometry_crs.isValid():
                df.setGeometrySrid(geometry_crs.postgisSrid())

        vLayer = QgsVectorLayer(df.toString(), "temp_vlayer", "virtual")
        if not vLayer.isValid():
            raise QgsProcessingException(vLayer.dataProvider().error().message())

        if vLayer.wkbType() == QgsWkbTypes.Type.Unknown:
            raise QgsProcessingException(self.tr("Cannot find geometry field"))

        (sink, dest_id) = self.parameterAsSink(
            parameters,
            self.OUTPUT,
            context,
            vLayer.fields(),
            vLayer.wkbType(),
            vLayer.crs(),
        )
        if sink is None:
            raise QgsProcessingException(self.invalidSinkError(parameters, self.OUTPUT))

        features = vLayer.getFeatures()
        total = 100.0 / vLayer.featureCount() if vLayer.featureCount() else 0
        for current, inFeat in enumerate(features):
            if feedback.isCanceled():
                break

            sink.addFeature(inFeat, QgsFeatureSink.Flag.FastInsert)
            feedback.setProgress(int(current * total))
        sink.finalize()
        return {self.OUTPUT: dest_id}



"""
***************************************************************************
    Heatmap.py
    ---------------------
    Date                 : December 2016
    Copyright            : (C) 2016 by Nyall Dawson
    Email                : nyall dot dawson at gmail dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Nyall Dawson"
__date__ = "December 2016"
__copyright__ = "(C) 2016, Nyall Dawson"

import os
from typing import Optional

from qgis.core import QgsProcessingUtils, QgsRectangle
from qgis.gui import QgsDoubleSpinBox
from qgis.PyQt import uic

from processing.gui.wrappers import DIALOG_STANDARD, WidgetWrapper
from processing.tools import dataobjects

pluginPath = os.path.dirname(__file__)
WIDGET, BASE = uic.loadUiType(os.path.join(pluginPath, "RasterResolutionWidget.ui"))


class HeatmapPixelSizeWidget(BASE, WIDGET):
    def __init__(self):
        super().__init__(None)
        self.setupUi(self)

        self.layer_bounds: QgsRectangle = QgsRectangle()
        self.source = None
        self.raster_bounds: QgsRectangle = QgsRectangle()
        self.radius: float = 100
        self.radius_field: Optional[str] = None

        self.mCellXSpinBox.setShowClearButton(False)
        self.mCellYSpinBox.setShowClearButton(False)
        self.mRowsSpinBox.setShowClearButton(False)
        self.mColumnsSpinBox.setShowClearButton(False)

        self.mCellYSpinBox.valueChanged.connect(self.mCellXSpinBox.setValue)
        self.mCellXSpinBox.valueChanged.connect(self.pixelSizeChanged)
        self.mRowsSpinBox.valueChanged.connect(self.rowsChanged)
        self.mColumnsSpinBox.valueChanged.connect(self.columnsChanged)

    def setRadius(self, radius: float):
        self.radius = radius
        self.recalculate_bounds()

    def setRadiusField(self, radius_field: Optional[str]):
        self.radius_field = radius_field
        self.recalculate_bounds()

    def setSource(self, source):
        if not source:
            return
        bounds = source.sourceExtent()
        if bounds.isNull():
            return

        self.source = source
        self.layer_bounds = bounds
        self.recalculate_bounds()

    def recalculate_bounds(self):
        self.raster_bounds = QgsRectangle(self.layer_bounds)

        if not self.source:
            return

        max_radius = self.radius
        if self.radius_field:
            idx = self.source.fields().lookupField(self.radius_field)
            try:
                max_radius = float(self.source.maximumValue(idx))
            except:
                pass

        self.raster_bounds.setXMinimum(self.raster_bounds.xMinimum() - max_radius)
        self.raster_bounds.setYMinimum(self.raster_bounds.yMinimum() - max_radius)
        self.raster_bounds.setXMaximum(self.raster_bounds.xMaximum() + max_radius)
        self.raster_bounds.setYMaximum(self.raster_bounds.yMaximum() + max_radius)

        self.pixelSizeChanged()

    def pixelSizeChanged(self):
        cell_size = self.mCellXSpinBox.value()
        if cell_size <= 0 or self.raster_bounds.isNull():
            return

        self.mCellYSpinBox.blockSignals(True)
        self.mCellYSpinBox.setValue(cell_size)
        self.mCellYSpinBox.blockSignals(False)
        rows = max(round(self.raster_bounds.height() / cell_size) + 1, 1)
        cols = max(round(self.raster_bounds.width() / cell_size) + 1, 1)
        self.mRowsSpinBox.blockSignals(True)
        self.mRowsSpinBox.setValue(rows)
        self.mRowsSpinBox.blockSignals(False)
        self.mColumnsSpinBox.blockSignals(True)
        self.mColumnsSpinBox.setValue(cols)
        self.mColumnsSpinBox.blockSignals(False)

    def rowsChanged(self):
        rows = self.mRowsSpinBox.value()
        if rows <= 0 or self.raster_bounds.isNull():
            return

        cell_size = self.raster_bounds.height() / rows
        if cell_size == 0:
            return

        cols = max(round(self.raster_bounds.width() / cell_size) + 1, 1)
        self.mColumnsSpinBox.blockSignals(True)
        self.mColumnsSpinBox.setValue(cols)
        self.mColumnsSpinBox.blockSignals(False)
        for w in [self.mCellXSpinBox, self.mCellYSpinBox]:
            w.blockSignals(True)
            w.setValue(cell_size)
            w.blockSignals(False)

    def columnsChanged(self):
        cols = self.mColumnsSpinBox.value()
        if cols < 2 or self.raster_bounds.isNull():
            return

        cell_size = self.raster_bounds.width() / (cols - 1)
        if cell_size == 0:
            return

        rows = max(round(self.raster_bounds.height() / cell_size), 1)
        self.mRowsSpinBox.blockSignals(True)
        self.mRowsSpinBox.setValue(rows)
        self.mRowsSpinBox.blockSignals(False)
        for w in [self.mCellXSpinBox, self.mCellYSpinBox]:
            w.blockSignals(True)
            w.setValue(cell_size)
            w.blockSignals(False)

    def setValue(self, value):
        try:
            numeric_value = float(value)
        except:
            return False

        self.mCellXSpinBox.setValue(numeric_value)
        self.mCellYSpinBox.setValue(numeric_value)
        return True

    def value(self):
        return self.mCellXSpinBox.value()


class HeatmapPixelSizeWidgetWrapper(WidgetWrapper):
    def __init__(self, param, dialog, row=0, col=0, **kwargs):
        super().__init__(param, dialog, row, col, **kwargs)
        self.context = dataobjects.createContext()

    def _panel(self):
        return HeatmapPixelSizeWidget()

    def createWidget(self):
        if self.dialogType == DIALOG_STANDARD:
            return self._panel()
        else:
            w = QgsDoubleSpinBox()
            w.setShowClearButton(False)
            w.setMinimum(0)
            w.setMaximum(99999999999)
            w.setDecimals(6)
            w.setToolTip(
                self.tr("Resolution of each pixel in output raster, in layer units")
            )
            return w

    def postInitialize(self, wrappers):
        if self.dialogType != DIALOG_STANDARD:
            return

        for wrapper in wrappers:
            if (
                wrapper.parameterDefinition().name()
                == self.parameterDefinition().parent_layer
            ):
                self.setSource(wrapper.parameterValue())
                wrapper.widgetValueHasChanged.connect(self.parentLayerChanged)
            elif (
                wrapper.parameterDefinition().name()
                == self.parameterDefinition().radius_param
            ):
                self.setRadius(wrapper.parameterValue())
                wrapper.widgetValueHasChanged.connect(self.radiusChanged)
            elif (
                wrapper.parameterDefinition().name()
                == self.parameterDefinition().radius_field_param
            ):
                self.setSource(wrapper.parameterValue())
                wrapper.widgetValueHasChanged.connect(self.radiusFieldChanged)

    def parentLayerChanged(self, wrapper):
        self.setSource(wrapper.parameterValue())

    def setSource(self, source):
        source = QgsProcessingUtils.variantToSource(source, self.context)
        self.widget.setSource(source)

    def radiusChanged(self, wrapper):
        self.setRadius(wrapper.parameterValue())

    def setRadius(self, radius: float):
        self.widget.setRadius(radius)

    def radiusFieldChanged(self, wrapper):
        self.setRadiusField(wrapper.parameterValue())

    def setRadiusField(self, radius_field: Optional[str]):
        self.widget.setRadiusField(radius_field)

    def setValue(self, value):
        return self.widget.setValue(value)

    def value(self):
        return self.widget.value()



"""
***************************************************************************
    Postprocessing.py
    ---------------------
    Date                 : August 2012
    Copyright            : (C) 2012 by Victor Olaya
    Email                : volayaf at gmail dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Victor Olaya"
__date__ = "August 2012"
__copyright__ = "(C) 2012, Victor Olaya"

import traceback
from typing import Optional

from qgis.core import (
    Qgis,
    QgsMapLayer,
    QgsMessageLog,
    QgsProcessingAlgorithm,
    QgsProcessingContext,
    QgsProcessingFeedback,
    QgsProcessingUtils,
    QgsWkbTypes,
)
from qgis.gui import QgsProcessingGuiUtils
from qgis.PyQt.QtCore import QCoreApplication
from qgis.utils import iface

from processing.core.ProcessingConfig import ProcessingConfig
from processing.gui.RenderingStyles import RenderingStyles


def determine_output_name(
    dest_id: str,
    details: QgsProcessingContext.LayerDetails,
    alg: QgsProcessingAlgorithm,
    context: QgsProcessingContext,
    parameters: dict,
) -> str:
    """
    If running a model, the execution will arrive here when an
    algorithm that is part of that model is executed. We check if
    its output is a final output of the model, and adapt the output
    name accordingly
    """
    for out in alg.outputDefinitions():
        if out.name() not in parameters:
            continue
        output_value = parameters[out.name()]
        if hasattr(output_value, "sink"):
            output_value = output_value.sink.valueAsString(context.expressionContext())[
                0
            ]
        else:
            output_value = str(output_value)
        if output_value == dest_id:
            return out.name()

    return details.outputName


def post_process_layer(
    output_name: str, layer: QgsMapLayer, alg: QgsProcessingAlgorithm
):
    """
    Applies post-processing steps to a layer
    """
    style = None
    if output_name:
        style = RenderingStyles.getStyle(alg.id(), output_name)

    if style is None:
        if layer.type() == Qgis.LayerType.Raster:
            style = ProcessingConfig.getSetting(ProcessingConfig.RASTER_STYLE)
        elif layer.type() == Qgis.LayerType.Vector:
            if layer.geometryType() == QgsWkbTypes.GeometryType.PointGeometry:
                style = ProcessingConfig.getSetting(ProcessingConfig.VECTOR_POINT_STYLE)
            elif layer.geometryType() == QgsWkbTypes.GeometryType.LineGeometry:
                style = ProcessingConfig.getSetting(ProcessingConfig.VECTOR_LINE_STYLE)
            else:
                style = ProcessingConfig.getSetting(
                    ProcessingConfig.VECTOR_POLYGON_STYLE
                )
    if style:
        layer.loadNamedStyle(style)

    if layer.type() == Qgis.LayerType.PointCloud:
        try:
            from qgis._3d import QgsPointCloudLayer3DRenderer

            if layer.renderer3D() is None:
                # If the layer has no 3D renderer and syncing 3D to 2D
                # renderer is enabled, we create a renderer and set it up
                # with the 2D renderer
                if layer.sync3DRendererTo2DRenderer():
                    renderer_3d = QgsPointCloudLayer3DRenderer()
                    renderer_3d.convertFrom2DRenderer(layer.renderer())
                    layer.setRenderer3D(renderer_3d)
        except ImportError:
            QgsMessageLog.logMessage(
                QCoreApplication.translate(
                    "Postprocessing",
                    "3D library is not available, "
                    "can't assign a 3d renderer to a layer.",
                )
            )


def handleAlgorithmResults(
    alg: QgsProcessingAlgorithm,
    context: QgsProcessingContext,
    feedback: Optional[QgsProcessingFeedback] = None,
    parameters: Optional[dict] = None,
):
    if not parameters:
        parameters = {}
    if feedback is None:
        feedback = QgsProcessingFeedback()
    wrong_layers = []

    feedback.setProgressText(
        QCoreApplication.translate("Postprocessing", "Loading resulting layers")
    )
    i = 0

    added_layers: list[QgsProcessingGuiUtils.ResultLayerDetails] = []
    layers_to_post_process: list[
        tuple[QgsMapLayer, QgsProcessingContext.LayerDetails]
    ] = []

    for dest_id, details in context.layersToLoadOnCompletion().items():
        if feedback.isCanceled():
            return False

        if len(context.layersToLoadOnCompletion()) > 2:
            # only show progress feedback if we're loading a bunch of layers
            feedback.setProgress(
                100 * i / float(len(context.layersToLoadOnCompletion()))
            )

        try:
            layer = QgsProcessingUtils.mapLayerFromString(
                dest_id, context, typeHint=details.layerTypeHint
            )
            if layer is not None:
                details.setOutputLayerName(layer)

                output_name = determine_output_name(
                    dest_id, details, alg, context, parameters
                )
                post_process_layer(output_name, layer, alg)

                # Load layer to layer tree root or to a specific group
                results_group = QgsProcessingGuiUtils.layerTreeResultsGroup(
                    details, context
                )

                # note here that we may not retrieve an owned layer -- eg if the
                # output layer already exists in the destination project
                owned_map_layer = context.temporaryLayerStore().takeMapLayer(layer)
                if owned_map_layer:
                    # we don't add the layer to the tree yet -- that's done
                    # later, after we've sorted all added layers
                    result_layer_details = QgsProcessingGuiUtils.ResultLayerDetails(
                        owned_map_layer
                    )
                    result_layer_details.targetLayerTreeGroup = results_group
                    result_layer_details.sortKey = details.layerSortKey or 0
                    result_layer_details.destinationProject = details.project
                    added_layers.append(result_layer_details)

                if details.postProcessor():
                    # we defer calling the postProcessor set in the context
                    # until the layer has been added to the project's layer
                    # tree, just in case the postProcessor contains logic
                    # relating to layer tree handling
                    layers_to_post_process.append((layer, details))

            else:
                wrong_layers.append(str(dest_id))
        except Exception:
            QgsMessageLog.logMessage(
                QCoreApplication.translate(
                    "Postprocessing", "Error loading result layer:"
                )
                + "\n"
                + traceback.format_exc(),
                "Processing",
                Qgis.MessageLevel.Critical,
            )
            wrong_layers.append(str(dest_id))
        i += 1

    if iface is not None:
        iface.layerTreeView().setUpdatesEnabled(False)

    QgsProcessingGuiUtils.addResultLayers(
        added_layers, context, iface.layerTreeView() if iface else None
    )

    # all layers have been added to the layer tree, so safe to call
    # postProcessors now
    for layer, details in layers_to_post_process:
        details.postProcessor().postProcessLayer(layer, context, feedback)

    if iface is not None:
        iface.layerTreeView().setUpdatesEnabled(True)

    feedback.setProgress(100)

    if wrong_layers:
        msg = QCoreApplication.translate(
            "Postprocessing", "The following layers were not correctly generated."
        )
        msg += "\n" + "\n".join([f"• {lay}" for lay in wrong_layers]) + "\n"
        msg += QCoreApplication.translate(
            "Postprocessing",
            "You can check the 'Log Messages Panel' in QGIS main window "
            "to find more information about the execution of the algorithm.",
        )
        feedback.reportError(msg)

    return len(wrong_layers) == 0



"""
***************************************************************************
    ClipRasterByExtent.py
    ---------------------
    Date                 : September 2013
    Copyright            : (C) 2013 by Alexander Bruy
    Email                : alexander bruy at gmail dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Alexander Bruy"
__date__ = "September 2013"
__copyright__ = "(C) 2013, Alexander Bruy"

import os

from qgis.core import (
    Qgis,
    QgsProcessingException,
    QgsProcessingParameterBoolean,
    QgsProcessingParameterDefinition,
    QgsProcessingParameterEnum,
    QgsProcessingParameterExtent,
    QgsProcessingParameterNumber,
    QgsProcessingParameterRasterDestination,
    QgsProcessingParameterRasterLayer,
    QgsProcessingParameterString,
    QgsProcessingRasterLayerDefinition,
    QgsRasterFileWriter,
)
from qgis.PyQt.QtGui import QIcon

from processing.algs.gdal.GdalAlgorithm import GdalAlgorithm
from processing.algs.gdal.GdalUtils import GdalUtils

pluginPath = os.path.split(os.path.split(os.path.dirname(__file__))[0])[0]


class ClipRasterByExtent(GdalAlgorithm):
    INPUT = "INPUT"
    EXTENT = "PROJWIN"
    OVERCRS = "OVERCRS"
    NODATA = "NODATA"
    OPTIONS = "OPTIONS"
    CREATION_OPTIONS = "CREATION_OPTIONS"
    DATA_TYPE = "DATA_TYPE"
    EXTRA = "EXTRA"
    OUTPUT = "OUTPUT"

    def __init__(self):
        super().__init__()

    def initAlgorithm(self, config=None):

        self.TYPES = [
            self.tr("Use Input Layer Data Type"),
            "Byte",
            "Int16",
            "UInt16",
            "UInt32",
            "Int32",
            "Float32",
            "Float64",
            "CInt16",
            "CInt32",
            "CFloat32",
            "CFloat64",
            "Int8",
        ]

        input_param = QgsProcessingParameterRasterLayer(
            self.INPUT, self.tr("Input layer")
        )
        # Support advance raster options panel
        input_param.setParameterCapabilities(
            input_param.parameterCapabilities()
            | Qgis.RasterProcessingParameterCapability.WmsScale
            | Qgis.RasterProcessingParameterCapability.WmsDpi
        )
        self.addParameter(input_param)

        self.addParameter(
            QgsProcessingParameterExtent(self.EXTENT, self.tr("Clipping extent"))
        )
        self.addParameter(
            QgsProcessingParameterBoolean(
                self.OVERCRS,
                self.tr("Override the projection for the output file"),
                defaultValue=False,
            )
        )
        self.addParameter(
            QgsProcessingParameterNumber(
                self.NODATA,
                self.tr("Assign a specified NoData value to output bands"),
                type=QgsProcessingParameterNumber.Type.Double,
                defaultValue=None,
                optional=True,
            )
        )

        # backwards compatibility parameter
        # TODO QGIS 5: remove parameter and related logic
        options_param = QgsProcessingParameterString(
            self.OPTIONS,
            self.tr("Additional creation options"),
            defaultValue="",
            optional=True,
        )
        options_param.setFlags(
            options_param.flags() | QgsProcessingParameterDefinition.Flag.Hidden
        )
        options_param.setMetadata({"widget_wrapper": {"widget_type": "rasteroptions"}})
        self.addParameter(options_param)

        creation_options_param = QgsProcessingParameterString(
            self.CREATION_OPTIONS,
            self.tr("Additional creation options"),
            defaultValue="",
            optional=True,
        )
        creation_options_param.setFlags(
            creation_options_param.flags()
            | QgsProcessingParameterDefinition.Flag.FlagAdvanced
        )
        creation_options_param.setMetadata(
            {"widget_wrapper": {"widget_type": "rasteroptions"}}
        )
        self.addParameter(creation_options_param)

        dataType_param = QgsProcessingParameterEnum(
            self.DATA_TYPE,
            self.tr("Output data type"),
            self.TYPES,
            allowMultiple=False,
            defaultValue=0,
        )
        dataType_param.setFlags(
            dataType_param.flags() | QgsProcessingParameterDefinition.Flag.FlagAdvanced
        )
        self.addParameter(dataType_param)

        extra_param = QgsProcessingParameterString(
            self.EXTRA,
            self.tr("Additional command-line parameters"),
            defaultValue=None,
            optional=True,
        )
        extra_param.setFlags(
            extra_param.flags() | QgsProcessingParameterDefinition.Flag.FlagAdvanced
        )
        self.addParameter(extra_param)

        self.addParameter(
            QgsProcessingParameterRasterDestination(
                self.OUTPUT, self.tr("Clipped (extent)")
            )
        )

    def name(self):
        return "cliprasterbyextent"

    def displayName(self):
        return self.tr("Clip raster by extent")

    def group(self):
        return self.tr("Raster extraction")

    def groupId(self):
        return "rasterextraction"

    def icon(self):
        return QIcon(os.path.join(pluginPath, "images", "gdaltools", "raster-clip.png"))

    def commandName(self):
        return "gdal_translate"

    def getConsoleCommands(self, parameters, context, feedback, executing=True):
        inLayer = self.parameterAsRasterLayer(parameters, self.INPUT, context)
        if inLayer is None:
            raise QgsProcessingException(
                "Invalid input layer {}".format(
                    parameters[self.INPUT] if self.INPUT in parameters else "INPUT"
                )
            )

        bbox = self.parameterAsExtent(parameters, self.EXTENT, context, inLayer.crs())
        if bbox.isNull() or bbox.isEmpty():
            raise QgsProcessingException(
                "Invalid extent {}".format(
                    parameters[self.EXTENT] if self.EXTENT in parameters else "EXTENT"
                )
            )

        input_details = GdalUtils.gdal_connection_details_from_layer(
            inLayer,
            self.INPUT,
            parameters,
            context,
            bbox,
        )

        arguments = []

        # If not a WMS or scale/DPI were not given, add -projwin
        if (
            inLayer.providerType() != "wms"
            or not isinstance(
                parameters[self.INPUT], QgsProcessingRasterLayerDefinition
            )
            or parameters[self.INPUT].referenceScale == 0
        ):
            arguments.extend(
                [
                    "-projwin",
                    str(bbox.xMinimum()),
                    str(bbox.yMaximum()),
                    str(bbox.xMaximum()),
                    str(bbox.yMinimum()),
                ]
            )

        override_crs = self.parameterAsBoolean(parameters, self.OVERCRS, context)
        if self.NODATA in parameters and parameters[self.NODATA] is not None:
            nodata = self.parameterAsDouble(parameters, self.NODATA, context)
        else:
            nodata = None
        out = self.parameterAsOutputLayer(parameters, self.OUTPUT, context)
        self.setOutputValue(self.OUTPUT, out)

        crs = inLayer.crs()
        if override_crs and crs.isValid():
            arguments.append(f"-a_srs {GdalUtils.gdal_crs_string(crs)}")

        if nodata is not None:
            arguments.append(f"-a_nodata {nodata}")

        data_type = self.parameterAsEnum(parameters, self.DATA_TYPE, context)
        if data_type:
            if self.TYPES[data_type] == "Int8" and GdalUtils.version() < 3070000:
                raise QgsProcessingException(
                    self.tr("Int8 data type requires GDAL version 3.7 or later")
                )

            arguments.append("-ot " + self.TYPES[data_type])

        output_format = self.outputFormat(parameters, self.OUTPUT, context)
        if not output_format:
            raise QgsProcessingException(self.tr("Output format is invalid"))

        arguments.append("-of")
        arguments.append(output_format)

        options = self.parameterAsString(parameters, self.CREATION_OPTIONS, context)
        # handle backwards compatibility parameter OPTIONS
        if self.OPTIONS in parameters and parameters[self.OPTIONS] not in (None, ""):
            options = self.parameterAsString(parameters, self.OPTIONS, context)
        if options:
            arguments.extend(GdalUtils.parseCreationOptions(options))

        if self.EXTRA in parameters and parameters[self.EXTRA] not in (None, ""):
            extra = self.parameterAsString(parameters, self.EXTRA, context)
            arguments.append(extra)

        arguments.append(input_details.connection_string)
        arguments.append(out)

        if input_details.open_options:
            arguments.extend(input_details.open_options_as_arguments())

        if input_details.credential_options:
            arguments.extend(input_details.credential_options_as_arguments())

        return [self.commandName(), GdalUtils.escapeAndJoin(arguments)]



"""
***************************************************************************
    RandomSelectionWithinSubsets.py
    ---------------------
    Date                 : August 2012
    Copyright            : (C) 2012 by Victor Olaya
    Email                : volayaf at gmail dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Victor Olaya"
__date__ = "August 2012"
__copyright__ = "(C) 2012, Victor Olaya"

import os
import random
from collections import defaultdict

from qgis.core import (
    QgsApplication,
    QgsFeatureRequest,
    QgsProcessingAlgorithm,
    QgsProcessingException,
    QgsProcessingOutputVectorLayer,
    QgsProcessingParameterEnum,
    QgsProcessingParameterFeatureSink,
    QgsProcessingParameterField,
    QgsProcessingParameterNumber,
    QgsProcessingParameterVectorLayer,
    QgsProcessingUtils,
)
from qgis.PyQt.QtGui import QIcon

from processing.algs.qgis.QgisAlgorithm import QgisAlgorithm

pluginPath = os.path.split(os.path.split(os.path.dirname(__file__))[0])[0]


class RandomSelectionWithinSubsets(QgisAlgorithm):
    INPUT = "INPUT"
    METHOD = "METHOD"
    NUMBER = "NUMBER"
    FIELD = "FIELD"
    OUTPUT = "OUTPUT"

    def icon(self):
        return QgsApplication.getThemeIcon("/algorithms/mAlgorithmSelectRandom.svg")

    def svgIconPath(self):
        return QgsApplication.iconPath("/algorithms/mAlgorithmSelectRandom.svg")

    def group(self):
        return self.tr("Vector selection")

    def groupId(self):
        return "vectorselection"

    def __init__(self):
        super().__init__()

    def flags(self):
        return (
            super().flags()
            | QgsProcessingAlgorithm.Flag.FlagNoThreading
            | QgsProcessingAlgorithm.Flag.FlagNotAvailableInStandaloneTool
        )

    def initAlgorithm(self, config=None):
        self.methods = [
            self.tr("Number of selected features"),
            self.tr("Percentage of selected features"),
        ]

        self.addParameter(
            QgsProcessingParameterVectorLayer(self.INPUT, self.tr("Input layer"))
        )
        self.addParameter(
            QgsProcessingParameterField(
                self.FIELD, self.tr("ID field"), None, self.INPUT
            )
        )
        self.addParameter(
            QgsProcessingParameterEnum(
                self.METHOD, self.tr("Method"), self.methods, False, 0
            )
        )
        self.addParameter(
            QgsProcessingParameterNumber(
                self.NUMBER,
                self.tr("Number/percentage of selected features"),
                QgsProcessingParameterNumber.Type.Integer,
                10,
                False,
                0.0,
            )
        )
        self.addOutput(
            QgsProcessingOutputVectorLayer(
                self.OUTPUT, self.tr("Selected (stratified random)")
            )
        )

    def name(self):
        return "randomselectionwithinsubsets"

    def displayName(self):
        return self.tr("Random selection within subsets")

    def shortDescription(self):
        return self.tr("Randomly selects features from a subset of a vector layer.")

    def shortHelpString(self):
        return self.tr(
            "This algorithm takes a vector layer and selects a subset of its features. "
            "No new layer is generated by this algorithm.\n"
            "The subset is defined randomly, using a percentage or count value to define "
            "the total number of features in the subset.\n"
            "The percentage/count value is not applied to the whole layer, but instead to each category. "
            "Categories are defined according to a given attribute, which is also specified "
            "as an input parameter for the algorithm."
        )

    def processAlgorithm(self, parameters, context, feedback):
        layer = self.parameterAsVectorLayer(parameters, self.INPUT, context)
        method = self.parameterAsEnum(parameters, self.METHOD, context)
        field = self.parameterAsString(parameters, self.FIELD, context)

        index = layer.fields().lookupField(field)

        unique = layer.uniqueValues(index)
        featureCount = layer.featureCount()

        value = self.parameterAsInt(parameters, self.NUMBER, context)
        if method == 0:
            if value > featureCount:
                raise QgsProcessingException(
                    self.tr(
                        "Selected number is greater that feature count. "
                        "Choose lesser value and try again."
                    )
                )
        else:
            if value > 100:
                raise QgsProcessingException(
                    self.tr(
                        "Percentage can't be greater than 100. Set a "
                        "different value and try again."
                    )
                )
            value = value / 100.0

        total = 100.0 / (featureCount * len(unique)) if featureCount else 1

        if len(unique) != featureCount:
            classes = defaultdict(list)

            features = layer.getFeatures(
                QgsFeatureRequest()
                .setFlags(QgsFeatureRequest.Flag.NoGeometry)
                .setSubsetOfAttributes([index])
            )

            for i, feature in enumerate(features):
                if feedback.isCanceled():
                    break

                classes[feature[index]].append(feature.id())
                feedback.setProgress(int(i * total))

            selran = []
            for k, subset in classes.items():
                if feedback.isCanceled():
                    break

                selValue = value if method != 1 else int(round(value * len(subset), 0))
                if selValue > len(subset):
                    selValue = len(subset)
                    feedback.reportError(
                        self.tr(
                            'Subset "{}" is smaller than requested number of features.'
                        ).format(k)
                    )
                selran.extend(random.sample(subset, selValue))

            layer.selectByIds(selran)
        else:
            layer.selectByIds(
                list(range(featureCount))
            )  # FIXME: implies continuous feature ids

        return {self.OUTPUT: parameters[self.INPUT]}



"""
***************************************************************************
    defaultproviders.py
    ---------------------
    Date                 : May 2016
    Copyright            : (C) 2016 by Victor Olaya
    Email                : volayaf at gmail dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Victor Olaya"
__date__ = "May 2016"
__copyright__ = "(C) 2016, Victor Olaya"


def loadDefaultProviders():
    # this is here just to "trigger" the above imports so providers are loaded
    # and can be found by the Processing.initialize() method
    pass



"""
/***************************************************************************
Python Console for QGIS
                             -------------------
begin                : 2012-09-10
copyright            : (C) 2012 by Salvatore Larosa
email                : lrssvtml (at) gmail (dot) com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
Some portions of code were taken from https://code.google.com/p/pydee/
"""

from __future__ import annotations

import code
import os
import re
import sys
import traceback
from functools import partial
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import TYPE_CHECKING, Optional

from qgis.core import (
    Qgis,
    QgsApplication,
    QgsProcessingUtils,
    QgsSettingsTree,
)
from qgis.gui import QgsCodeEditor, QgsCodeEditorPython, QgsCodeInterpreter
from qgis.PyQt.Qsci import QsciScintilla
from qgis.PyQt.QtCore import QCoreApplication, Qt
from qgis.PyQt.QtGui import QClipboard, QCursor, QFontMetrics, QKeySequence
from qgis.PyQt.QtWidgets import QAction, QApplication, QShortcut

from .process_wrapper import ProcessWrapper

if TYPE_CHECKING:
    from .console import PythonConsoleWidget

_init_statements = [
    # Python
    "import sys",
    "import os",
    "from pathlib import Path",
    "import re",
    "import math",
    # QGIS
    "from qgis.core import *",
    "from qgis.gui import *",
    "from qgis.analysis import *",
    # # 3D might not be compiled in
    """
try:
    from qgis._3d import *
except ModuleNotFoundError:
    pass
""",
    "import processing",
    "import qgis.utils",
    "from qgis.utils import iface",
    # Qt
    "from qgis.PyQt.QtCore import *",
    "from qgis.PyQt.QtGui import *",
    "from qgis.PyQt.QtWidgets import *",
    "from qgis.PyQt.QtNetwork import *",
    "from qgis.PyQt.QtXml import *",
    r"""
def __parse_object(object=None):
    if not object:
        return None
    import inspect
    if inspect.isclass(object):
        str_class = str(object)
    else:
        str_class = str(object.__class__)

    qgis_api_pattern = r".*qgis\._(\w+)\.(\w+).*"
    match = re.match(qgis_api_pattern, str_class)
    if match:
        module = match[1]
        obj = match[2]
        return 'qgis', module, obj

    pyqt_pattern = r".*PyQt5\.(\w+)\.(\w+).*"
    match = re.match(pyqt_pattern, str_class)
    if match:
        module = match[1]
        obj = match[2]
        return 'qt', module, obj
""",
    r"""
def _help(object=None, api=Qgis.DocumentationApi.PyQgis, force_search=False):
    '''
    Link to the C++ or PyQGIS API documentation for the given object.
    If no object is given, the main PyQGIS API page is opened.
    If the object is not part of the QGIS API but is a Qt object the Qt documentation is opened.
    '''

    pythonSettingsTreeNode = QgsSettingsTree.node("gui").childNode("code-editor").childNode("python")
    browserName = pythonSettingsTreeNode.childSetting('context-help-browser').valueAsVariant()
    try:
        browser = Qgis.DocumentationBrowser[browserName]
    except KeyError:
        browser = Qgis.DocumentationBrowser.DeveloperToolsPanel

    if not object:
        return iface.showApiDocumentation(api, browser=browser)

    def search_or_home(object_str):
        if not object_str:
            return iface.showApiDocumentation(api, browser=browser)
        if browser == Qgis.DocumentationBrowser.DeveloperToolsPanel and not QgsGui.hasWebEngine():
            if force_search:
                return iface.showApiDocumentation(Qgis.DocumentationApi.PyQgisSearch, object=object, browser=Qgis.DocumentationBrowser.SystemWebBrowser)
            else:
                return iface.showApiDocumentation(api, browser=browser)
        else:
            return iface.showApiDocumentation(Qgis.DocumentationApi.PyQgisSearch, object=object, browser=browser)

    if isinstance(object, str):
        try:
            object = eval(object)
        except (SyntaxError, NameError):
            return search_or_home(object)

    obj_info = __parse_object(object)
    if not obj_info:
        return search_or_home(object if isinstance(object, str) else None)

    obj_type, module, class_name = obj_info
    if obj_type == "qt":
        api = Qgis.DocumentationApi.Qt

    iface.showApiDocumentation(api, browser=browser, object=class_name, module=module)

""",
    r"""
def _api(object=None):
    '''
    Link to the QGIS API documentation for the given object.
    If no object is given, the main API page is opened.
    If the object is not part of the QGIS API but is a Qt object the Qt documentation is opened.
    '''
    return _help(object, api=Qgis.DocumentationApi.CppQgis)
""",
    r"""
def _pyqgis(object=None):
    '''
    Link to the PyQGIS API documentation for the given object.
    If no object is given, the main PyQGIS API page is opened.
    If the object is not part of the QGIS API but is a Qt object the Qt documentation is opened.
    '''
    return _help(object, api=Qgis.DocumentationApi.PyQgis)
""",
]


# States of the interpreter
PS1 = 0  # Writing a new command
PS2 = 1  # Continuation of a multi-line command
SUBPROCESS = 2  # Sending input to a subprocess


class PythonInterpreter(QgsCodeInterpreter, code.InteractiveInterpreter):
    def __init__(self, shell: ShellScintilla):
        super(QgsCodeInterpreter, self).__init__()
        code.InteractiveInterpreter.__init__(self, locals=None)

        self.shell: ShellScintilla = shell
        self.sub_process = None
        self.buffer = []

        for statement in _init_statements:
            try:
                self.runsource(statement)
            except ModuleNotFoundError:
                pass

    def execCommandImpl(self, cmd, show_input=True):

        # Child process running, input should be sent to it
        if self.currentState() == SUBPROCESS:
            sys.stdout.write(cmd + "\n")
            self.sub_process.write(cmd)
            return 0

        if show_input:
            self.writeCMD(cmd)

        if self.currentState() == PS1:
            # This line makes single line commands with leading spaces work
            cmd = cmd.strip()

            # User entered: varname = !cmd
            # Run the command and assign the output to varname
            # Mimics IPython's behavior
            assignment_pattern = r"(\w+)\s*=\s*!(.+)"
            match = re.match(assignment_pattern, cmd)
            if match:
                varname = match[1]
                cmd = match[2]
                # Run the command in non-interactive mode
                self.sub_process = ProcessWrapper(cmd, interactive=False)
                # Concatenate stdout and stderr
                res = (self.sub_process.stdout + self.sub_process.stderr).strip()

                # Use a temporary file to communicate the result to the inner interpreter
                with NamedTemporaryFile(encoding="utf-8", mode="w", delete=False) as f:
                    name = f.name
                    f.write(res)
                tmp = Path(name)
                self.runsource(
                    f'{varname} = Path("{tmp}").read_text(encoding="utf-8").split("\\n")'
                )
                tmp.unlink()
                self.sub_process = None
                return 0

            # User entered: !cmd
            # Run the command and stream the output to the console
            # While the process is running, the console is in state 2 meaning
            # that all input is sent to the child process
            # Mimics IPython's behavior
            elif cmd.startswith("!"):
                cmd = cmd[1:]
                self.sub_process = ProcessWrapper(cmd)
                self.sub_process.finished.connect(self.processFinished)
                return 0

        res = 0

        import webbrowser

        version = (
            "master"
            if "master" in Qgis.QGIS_VERSION.lower()
            else re.findall(r"^\d.[0-9]*", Qgis.QGIS_VERSION)[0]
        )

        if cmd == "?":
            self.shell.console_widget.shell_output.insertHelp()
        elif cmd == "_pyqgis":
            self.shell.showApi(Qgis.DocumentationApi.PyQgis)
        elif cmd == "_api":
            self.shell.showApi(Qgis.DocumentationApi.CppQgis)
        elif cmd == "_cookbook":
            webbrowser.open(
                "https://docs.qgis.org/{}/en/docs/pyqgis_developer_cookbook/".format(
                    "testing" if version == "master" else version
                )
            )
        else:
            self.buffer.append(cmd)
            src = "\n".join(self.buffer)
            res = self.runsource(src)
            if res == 0:
                self.buffer = []

        return res

    def writeCMD(self, txt):
        if sys.stdout:
            sys.stdout.fire_keyboard_interrupt = False
        if len(txt) > 0:
            sys.stdout.write(f"{self.promptForState()} {txt}\n")

    def runsource(self, source, filename="<input>", symbol="single"):
        if sys.stdout:
            sys.stdout.fire_keyboard_interrupt = False

        hook = sys.excepthook
        try:

            def excepthook(etype, value, tb):
                self.write("".join(traceback.format_exception(etype, value, tb)))

            sys.excepthook = excepthook

            return super().runsource(source, filename, symbol)
        finally:
            sys.excepthook = hook

    def currentState(self):
        if self.sub_process:
            return SUBPROCESS
        return super().currentState()

    def promptForState(self, state=-1):
        if state == -1:
            state = self.currentState()
        if state == SUBPROCESS:
            return " : "
        elif state == PS2:
            return "..."
        else:
            return ">>>"

    def processFinished(self, errorcode):
        self.sub_process = None
        self.shell.updatePrompt()


class ShellScintilla(QgsCodeEditorPython):
    def __init__(self, console_widget: PythonConsoleWidget):
        # We set the ImmediatelyUpdateHistory flag here, as users can easily
        # crash QGIS by entering a Python command, and we don't want the
        # history leading to the crash lost...
        super().__init__(
            console_widget,
            [],
            QgsCodeEditor.Mode.CommandInput,
            flags=QgsCodeEditor.Flags(
                QgsCodeEditor.Flag.CodeFolding
                | QgsCodeEditor.Flag.ImmediatelyUpdateHistory
            ),
        )

        self.console_widget: PythonConsoleWidget = console_widget
        self._interpreter = PythonInterpreter(shell=self)
        self.setInterpreter(self._interpreter)

        self.opening = ["(", "{", "[", "'", '"']
        self.closing = [")", "}", "]", "'", '"']

        self.setHistoryFilePath(
            os.path.join(QgsApplication.qgisSettingsDirPath(), "console_history.txt")
        )

        self.refreshSettingsShell()

        # Disable command key
        ctrl, shift = self.SCMOD_CTRL << 16, self.SCMOD_SHIFT << 16
        self.SendScintilla(QsciScintilla.SCI_CLEARCMDKEY, ord("L") + ctrl)
        self.SendScintilla(QsciScintilla.SCI_CLEARCMDKEY, ord("T") + ctrl)
        self.SendScintilla(QsciScintilla.SCI_CLEARCMDKEY, ord("D") + ctrl)
        self.SendScintilla(QsciScintilla.SCI_CLEARCMDKEY, ord("Z") + ctrl)
        self.SendScintilla(QsciScintilla.SCI_CLEARCMDKEY, ord("Y") + ctrl)
        self.SendScintilla(QsciScintilla.SCI_CLEARCMDKEY, ord("L") + ctrl + shift)
        self.SendScintilla(QsciScintilla.SCI_CLEARCMDKEY, ord("V") + ctrl)

        # New QShortcut = ctrl+space/ctrl+alt+space for Autocomplete
        self.newShortcutCSS = QShortcut(
            QKeySequence(Qt.Modifier.CTRL | Qt.Modifier.SHIFT | Qt.Key.Key_Space), self
        )
        self.newShortcutCAS = QShortcut(
            QKeySequence(Qt.Modifier.CTRL | Qt.Modifier.ALT | Qt.Key.Key_Space), self
        )
        self.newShortcutCSS.setContext(Qt.ShortcutContext.WidgetShortcut)
        self.newShortcutCAS.setContext(Qt.ShortcutContext.WidgetShortcut)
        self.newShortcutCAS.activated.connect(self.autoComplete)
        self.newShortcutCSS.activated.connect(self.showHistory)

        self.sessionHistoryCleared.connect(self.on_session_history_cleared)
        self.persistentHistoryCleared.connect(self.on_persistent_history_cleared)

    def _setMinimumHeight(self):
        font = self.lexer().defaultFont(0)
        fm = QFontMetrics(font)

        self.setMinimumHeight(fm.height() + 10)

    def refreshSettingsShell(self):
        # Set Python lexer
        self.initializeLexer()

        # Sets minimum height for input area based of font metric
        self._setMinimumHeight()

    def on_session_history_cleared(self):
        msgText = QCoreApplication.translate(
            "PythonConsole", "Session history cleared successfully."
        )
        self.console_widget.callWidgetMessageBar(msgText)

    def on_persistent_history_cleared(self):
        msgText = QCoreApplication.translate(
            "PythonConsole", "History cleared successfully."
        )
        self.console_widget.callWidgetMessageBar(msgText)

    def keyPressEvent(self, e):
        if (
            e.modifiers()
            & (Qt.KeyboardModifier.ControlModifier | Qt.KeyboardModifier.MetaModifier)
            and e.key() == Qt.Key.Key_C
            and not self.hasSelectedText()
        ):
            if self._interpreter.sub_process:
                sys.stderr.write("Terminate child process\n")
                self._interpreter.sub_process.kill()
                self._interpreter.sub_process = None
                self.updatePrompt()
            return
        elif (
            e.modifiers()
            & (Qt.KeyboardModifier.ControlModifier | Qt.KeyboardModifier.MetaModifier)
            and e.key() == Qt.Key.Key_V
        ) or (
            e.modifiers() & Qt.KeyboardModifier.ShiftModifier
            and e.key() == Qt.Key.Key_Insert
        ):
            self.paste()
            e.accept()
            return

        # update the live history
        self.updateSoftHistory()
        super().keyPressEvent(e)
        self.updatePrompt()

    def mousePressEvent(self, e):
        """
        Re-implemented to handle the mouse press event.
        e: the mouse press event (QMouseEvent)
        """
        self.setFocus()
        if e.button() == Qt.MouseButton.MiddleButton:
            stringSel = QApplication.clipboard().text(QClipboard.Mode.Selection)
            if not self.isCursorOnLastLine():
                self.moveCursorToEnd()
            self.insertFromDropPaste(stringSel)
            e.accept()
        else:
            QgsCodeEditorPython.mousePressEvent(self, e)

    def paste(self):
        """
        Method to display data from the clipboard.

        XXX: It should reimplement the virtual QScintilla.paste method,
        but it seems not used by QScintilla code.
        """
        stringPaste = QApplication.clipboard().text()
        if self.isCursorOnLastLine():
            if self.hasSelectedText():
                self.removeSelectedText()
        else:
            self.moveCursorToEnd()
        self.insertFromDropPaste(stringPaste)

    # Drag and drop
    def dropEvent(self, e):
        if e.mimeData().hasText():
            stringDrag = e.mimeData().text()
            self.insertFromDropPaste(stringDrag)
            self.setFocus()
            e.setDropAction(Qt.DropAction.CopyAction)
            e.accept()
        else:
            QgsCodeEditorPython.dropEvent(self, e)

    def insertFromDropPaste(self, textDP):
        pasteList = textDP.splitlines()
        if pasteList:
            for line in pasteList[:-1]:
                cleanLine = line.replace(">>> ", "").replace("... ", "")
                self.insert(cleanLine)
                self.moveCursorToEnd()
                self.runCommand(self.text())
            if pasteList[-1] != "":
                line = pasteList[-1]
                cleanLine = line.replace(">>> ", "").replace("... ", "")
                curpos = self.getCursorPosition()
                self.insert(cleanLine)
                self.setCursorPosition(curpos[0], curpos[1] + len(cleanLine))

    def insertTextFromFile(self, listOpenFile):
        for line in listOpenFile[:-1]:
            self.append(line)
            self.moveCursorToEnd()
            self.SendScintilla(QsciScintilla.SCI_DELETEBACK)
            self.runCommand(self.text())
        self.append(listOpenFile[-1])
        self.moveCursorToEnd()
        self.SendScintilla(QsciScintilla.SCI_DELETEBACK)

    def entered(self):
        self.moveCursorToEnd()
        self.runCommand(self.text())
        self.setFocus()
        self.moveCursorToEnd()

    def write(self, txt):
        if sys.stderr:
            sys.stderr.write(txt)

    def runFile(self, filename, override_file_name: str | None = None):
        filename = filename.replace("\\", "/")
        dirname = os.path.dirname(filename)

        # Append the directory of the file to the path and set __file__ to the filename
        self._interpreter.execCommandImpl(
            f"sys.path.append({QgsProcessingUtils.stringToPythonLiteral(dirname)})",
            False,
        )
        self._interpreter.execCommandImpl(
            f"__file__ = {QgsProcessingUtils.stringToPythonLiteral(filename)}",
            False,
        )

        try:
            # Run the file

            self.runCommand(
                "exec(compile(Path({}).read_text(), {}, 'exec'))".format(
                    QgsProcessingUtils.stringToPythonLiteral(filename),
                    QgsProcessingUtils.stringToPythonLiteral(
                        override_file_name or filename
                    ),
                ),
                skipHistory=True,
            )
        finally:
            # Remove the directory from the path and delete the __file__ variable
            self._interpreter.execCommandImpl("del __file__", False)
            self._interpreter.execCommandImpl(
                f"sys.path.remove({QgsProcessingUtils.stringToPythonLiteral(dirname)})",
                False,
            )

    def showApiDocumentation(self, text, force_search=False):
        self._interpreter.execCommandImpl(
            f"_help({repr(text)}, api=Qgis.DocumentationApi.PyQgis, force_search={force_search})",
            show_input=False,
        )

    def showApi(self, api: Qgis.DocumentationApi):
        self._interpreter.execCommandImpl(
            f"_help(api=Qgis.DocumentationApi.{api.name})", show_input=False
        )

    def populateContextMenu(self, menu):

        word = self.selectedText() or self.wordAtPoint(
            self.mapFromGlobal(QCursor.pos())
        )
        if word:
            context_help_action = QAction(
                QgsApplication.getThemeIcon("mActionHelpContents.svg"),
                QCoreApplication.translate("PythonConsole", "Context Help"),
                menu,
            )
            context_help_action.triggered.connect(
                partial(self.showApiDocumentation, word, force_search=True)
            )
            context_help_action.setShortcut(QKeySequence.StandardKey.HelpContents)
            menu.addAction(context_help_action)



"""
***************************************************************************
    LinesToPolygons.py
    ---------------------
    Date                 : August 2012
    Copyright            : (C) 2012 by Victor Olaya
    Email                : volayaf at gmail dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Victor Olaya"
__date__ = "August 2012"
__copyright__ = "(C) 2012, Victor Olaya"

import os

from qgis.core import (
    QgsApplication,
    QgsFeature,
    QgsFeatureSink,
    QgsGeometry,
    QgsGeometryCollection,
    QgsMultiPolygon,
    QgsMultiSurface,
    QgsPolygon,
    QgsProcessing,
    QgsProcessingParameterFeatureSink,
    QgsProcessingParameterFeatureSource,
    QgsProcessingUtils,
    QgsWkbTypes,
)
from qgis.PyQt.QtGui import QIcon

from processing.algs.qgis.QgisAlgorithm import QgisFeatureBasedAlgorithm
from processing.tools import dataobjects, vector

pluginPath = os.path.split(os.path.split(os.path.dirname(__file__))[0])[0]


class LinesToPolygons(QgisFeatureBasedAlgorithm):
    def icon(self):
        return QgsApplication.getThemeIcon("/algorithms/mAlgorithmLineToPolygon.svg")

    def svgIconPath(self):
        return QgsApplication.iconPath("/algorithms/mAlgorithmLineToPolygon.svg")

    def tags(self):
        return self.tr("line,polygon,convert").split(",")

    def group(self):
        return self.tr("Vector geometry")

    def groupId(self):
        return "vectorgeometry"

    def __init__(self):
        super().__init__()

    def name(self):
        return "linestopolygons"

    def displayName(self):
        return self.tr("Lines to polygons")

    def shortDescription(self):
        return self.tr("Generates polygons using the lines from an input line layer.")

    def shortHelpString(self):
        return self.tr(
            "This algorithm generates a polygon layer using as polygon rings the lines from an input line layer.\n"
            "All attributes from the input line layer will be copied to the output layer."
        )

    def outputName(self):
        return self.tr("Polygons")

    def outputType(self):
        return QgsProcessing.SourceType.TypeVectorPolygon

    def inputLayerTypes(self):
        return [QgsProcessing.SourceType.TypeVectorLine]

    def outputWkbType(self, input_wkb_type):
        return self.convertWkbToPolygons(input_wkb_type)

    def processFeature(self, feature, context, feedback):
        if feature.hasGeometry():
            feature.setGeometry(QgsGeometry(self.convertToPolygons(feature.geometry())))
            if feature.geometry().isEmpty():
                feedback.reportError(
                    self.tr(
                        "One or more line ignored due to geometry not having a minimum of three vertices."
                    )
                )
        return [feature]

    def supportInPlaceEdit(self, layer):
        return False

    def convertWkbToPolygons(self, wkb):
        multi_wkb = QgsWkbTypes.Type.NoGeometry
        if (
            QgsWkbTypes.singleType(QgsWkbTypes.flatType(wkb))
            == QgsWkbTypes.Type.LineString
        ):
            multi_wkb = QgsWkbTypes.Type.MultiPolygon
        elif (
            QgsWkbTypes.singleType(QgsWkbTypes.flatType(wkb))
            == QgsWkbTypes.Type.CompoundCurve
        ):
            multi_wkb = QgsWkbTypes.Type.MultiSurface
        if QgsWkbTypes.hasM(wkb):
            multi_wkb = QgsWkbTypes.addM(multi_wkb)
        if QgsWkbTypes.hasZ(wkb):
            multi_wkb = QgsWkbTypes.addZ(multi_wkb)

        return multi_wkb

    def convertToPolygons(self, geometry):
        surfaces = self.getSurfaces(geometry.constGet())
        output_wkb = self.convertWkbToPolygons(geometry.wkbType())
        out_geom = None
        if QgsWkbTypes.flatType(output_wkb) == QgsWkbTypes.Type.MultiPolygon:
            out_geom = QgsMultiPolygon()
        else:
            out_geom = QgsMultiSurface()

        for surface in surfaces:
            out_geom.addGeometry(surface)

        return out_geom

    def getSurfaces(self, geometry):
        surfaces = []
        if isinstance(geometry, QgsGeometryCollection):
            # collection
            for i in range(geometry.numGeometries()):
                surfaces.extend(self.getSurfaces(geometry.geometryN(i)))
        else:
            # not collection
            if geometry.vertexCount() > 2:
                surface = QgsPolygon()
                surface.setExteriorRing(geometry.clone())
                surfaces.append(surface)

        return surfaces



"""
***************************************************************************
    algfactory.py
    ---------------------
    Date                 : November 2018
    Copyright            : (C) 2018 by Nathan Woodrow
    Email                : woodrow dot nathan at gmail dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Nathan Woodrow"
__date__ = "November 2018"
__copyright__ = "(C) 2018, Nathan Woodrow"

from collections import OrderedDict
from functools import partial

from qgis.core import (
    QgsApplication,
    QgsMessageLog,
    QgsProcessingAlgorithm,
    QgsProcessingOutputBoolean,
    QgsProcessingOutputFile,
    QgsProcessingOutputFolder,
    QgsProcessingOutputHtml,
    QgsProcessingOutputLayerDefinition,
    QgsProcessingOutputMapLayer,
    QgsProcessingOutputMultipleLayers,
    QgsProcessingOutputNumber,
    QgsProcessingOutputPointCloudLayer,
    QgsProcessingOutputRasterLayer,
    QgsProcessingOutputString,
    QgsProcessingOutputVectorLayer,
    QgsProcessingParameterAnnotationLayer,
    QgsProcessingParameterAuthConfig,
    QgsProcessingParameterBand,
    QgsProcessingParameterBoolean,
    QgsProcessingParameterColor,
    QgsProcessingParameterCoordinateOperation,
    QgsProcessingParameterCrs,
    QgsProcessingParameterDatabaseSchema,
    QgsProcessingParameterDatabaseTable,
    QgsProcessingParameterDateTime,
    QgsProcessingParameterDefinition,
    QgsProcessingParameterDistance,
    QgsProcessingParameterDuration,
    QgsProcessingParameterEnum,
    QgsProcessingParameterExpression,
    QgsProcessingParameterExtent,
    QgsProcessingParameterFeatureSink,
    QgsProcessingParameterFeatureSource,
    QgsProcessingParameterField,
    QgsProcessingParameterFile,
    QgsProcessingParameterFileDestination,
    QgsProcessingParameterFolderDestination,
    QgsProcessingParameterGeometry,
    QgsProcessingParameterLayout,
    QgsProcessingParameterLayoutItem,
    QgsProcessingParameterMapLayer,
    QgsProcessingParameterMapTheme,
    QgsProcessingParameterMatrix,
    QgsProcessingParameterMeshLayer,
    QgsProcessingParameterMultipleLayers,
    QgsProcessingParameterNumber,
    QgsProcessingParameterPoint,
    QgsProcessingParameterPointCloudDestination,
    QgsProcessingParameterPointCloudLayer,
    QgsProcessingParameterProviderConnection,
    QgsProcessingParameterRange,
    QgsProcessingParameterRasterDestination,
    QgsProcessingParameterRasterLayer,
    QgsProcessingParameterScale,
    QgsProcessingParameterString,
    QgsProcessingParameterVectorDestination,
    QgsProcessingParameterVectorLayer,
)
from qgis.PyQt.QtCore import QCoreApplication
from qgis.PyQt.QtGui import QIcon


def _log(*args, **kw):
    """
    Log messages to the QgsMessageLog viewer
    """
    QgsMessageLog.logMessage(" ".join(map(str, args)), "Factory")


def _make_output(**args):
    """
    Create a processing output class type.
    :param args: 'cls' The class object type.
                 'name' the name of the output
                 'description' The description used on the output
    :return:
    """
    cls = args["cls"]
    del args["cls"]
    newargs = {
        "name": args["name"],
        "description": args["description"],
    }
    return cls(**newargs)


class ProcessingAlgFactoryException(Exception):
    """
    Exception raised when using @alg on a function.
    """

    def __init__(self, message):
        super().__init__(message)


class AlgWrapper(QgsProcessingAlgorithm):
    """
    Wrapper object used to create new processing algorithms from @alg.
    """

    def __init__(
        self,
        name=None,
        display=None,
        group=None,
        group_id=None,
        inputs=None,
        outputs=None,
        func=None,
        help=None,
        icon=None,
    ):
        super().__init__()
        self._inputs = OrderedDict(inputs or {})
        self._outputs = OrderedDict(outputs or {})
        self._icon = icon
        self._name = name
        self._group = group
        self._group_id = group_id
        self._display = display
        self._func = func
        self._help = help

    def _get_parent_id(self, parent):
        """
        Return the id of the parent object.
        """
        if isinstance(parent, str):
            return parent
        else:
            raise NotImplementedError()

    # Wrapper logic
    def define(
        self,
        name,
        label,
        group,
        group_label,
        help=None,
        icon=QgsApplication.iconPath("processingScript.svg"),
    ):
        self._name = name
        self._display = label
        self._group = group_label
        self._group_id = group
        self._help = help
        self._icon = icon

    def end(self):
        """
        Finalize the wrapper logic and check for any invalid config.
        """
        if not self.has_outputs:
            raise ProcessingAlgFactoryException(
                "No outputs defined for '{}' alg"
                "At least one must be defined. Use @alg.output"
            )

    def add_output(self, type, **kwargs):
        parm = self._create_param(type, output=True, **kwargs)
        self._outputs[parm.name()] = parm

    def add_help(self, helpstring, *args, **kwargs):
        self._help = helpstring

    def add_input(self, type, **kwargs):
        parm = self._create_param(type, **kwargs)
        self._inputs[parm.name()] = parm

    @property
    def inputs(self):
        return self._inputs

    @property
    def outputs(self):
        return self._outputs

    def _create_param(self, type, output=False, **kwargs):
        name = kwargs["name"]
        if name in self._inputs or name in self._outputs:
            raise ProcessingAlgFactoryException(f"{name} already defined")

        parent = kwargs.get("parent")
        if parent:
            parentname = self._get_parent_id(parent)
            if parentname == name:
                raise ProcessingAlgFactoryException(
                    f"{name} can't depend on itself. "
                    "We know QGIS is smart but it's not that smart"
                )
            if parentname not in self._inputs and parentname not in self._outputs:
                raise ProcessingAlgFactoryException(
                    f"Can't find parent named {parentname}"
                )

        kwargs["description"] = kwargs.pop("label", "")
        kwargs["defaultValue"] = kwargs.pop("default", None)
        advanced = kwargs.pop("advanced", False)
        help_str = kwargs.pop("help", "")
        try:
            if output:
                try:
                    make_func = output_type_mapping[type]
                except KeyError:
                    raise ProcessingAlgFactoryException(
                        f"{type} is a invalid output type"
                    )
            else:
                try:
                    make_func = input_type_mapping[type]
                except KeyError:
                    raise ProcessingAlgFactoryException(
                        f"{type} is a invalid input type"
                    )
            parm = make_func(**kwargs)
            if advanced:
                parm.setFlags(
                    parm.flags() | QgsProcessingParameterDefinition.Flag.FlagAdvanced
                )
            if not output:
                parm.setHelp(help_str)
            return parm
        except KeyError as ex:
            raise NotImplementedError(f"{str(type)} not supported")

    def set_func(self, func):
        self._func = func
        # Only take the help from the function if it hasn't already been set.
        if self._func and not self._help:
            self._help = self._func.__doc__.strip()

    @property
    def has_outputs(self):
        """
        True if this alg wrapper has outputs defined.
        """
        dests = [p for p in self._inputs.values() if p.isDestination()]
        return bool(self._outputs) or bool(dests)

    @property
    def has_inputs(self):
        """
        True if this alg wrapper has outputs defined.
        """
        return bool(self._inputs)

    def _get_parameter_value(self, parm, parameters, name, context):
        """
        Extract the real value from the parameter.
        """
        if isinstance(
            parm, (QgsProcessingParameterString, QgsProcessingParameterAuthConfig)
        ):
            value = self.parameterAsString(parameters, name, context)
            return value
        elif isinstance(parm, QgsProcessingParameterNumber):
            if parm.dataType() == QgsProcessingParameterNumber.Type.Integer:
                value = self.parameterAsInt(parameters, name, context)
                return value
            if parm.dataType() == QgsProcessingParameterNumber.Type.Double:
                value = self.parameterAsDouble(parameters, name, context)
                return value

    # Overloads
    def name(self):
        return self._name

    def displayName(self):
        return self._display

    def group(self):
        return self._group

    def groupId(self):
        return self._group_id

    def processAlgorithm(self, parameters, context, feedback):
        values = {}
        for parm in self._inputs.values():
            name = parm.name()
            values[name] = self._get_parameter_value(parm, parameters, name, context)

        output = self._func(self, parameters, context, feedback, values)
        if output is None:
            return {}
        return output

    def createInstance(self):
        return AlgWrapper(
            self._name,
            self._display,
            self._group,
            self._group_id,
            inputs=self._inputs,
            outputs=self._outputs,
            func=self._func,
            help=self._help,
            icon=self._icon,
        )

    def initAlgorithm(self, configuration=None, p_str=None, Any=None, *args, **kwargs):
        for parm in self._inputs.values():
            self.addParameter(parm.clone())

        for parm in self._outputs.values():
            clsname = parm.__class__.__name__
            klass = globals()[clsname]
            clone = klass(parm.name(), parm.description())
            self.addOutput(clone)

    def shortHelpString(self):
        return self._help

    def icon(self):
        return QIcon(self._icon)


class ProcessingAlgFactory:
    STRING = ("STRING",)
    INT = ("INT",)
    NUMBER = ("NUMBER",)
    DISTANCE = ("DISTANCE",)
    SINK = "SINK"
    SOURCE = "SOURCE"
    FILE = ("FILE",)
    FOLDER = ("FOLDER",)
    HTML = ("HTML",)
    LAYERDEF = ("LAYERDEF",)
    MAPLAYER = ("MAPLAYER",)
    MULTILAYER = ("MULTILAYER",)
    RASTER_LAYER = ("RASTER_LAYER",)
    VECTOR_LAYER = ("VECTOR_LAYER",)
    MESH_LAYER = ("MESH_LAYER",)
    POINT_CLOUD_LAYER = ("POINT_CLOUD_LAYER",)
    FILE_DEST = ("FILE_DEST",)
    FOLDER_DEST = ("FOLDER_DEST",)
    RASTER_LAYER_DEST = ("RASTER_LAYER_DEST",)
    VECTOR_LAYER_DEST = ("VECTOR_LAYER_DEST",)
    POINTCLOUD_LAYER_DEST = ("POINTCLOUD_LAYER_DEST",)
    BAND = ("BAND",)
    BOOL = ("BOOL",)
    CRS = ("CRS",)
    ENUM = ("ENUM",)
    EXPRESSION = ("EXPRESSION",)
    EXTENT = ("EXTENT",)
    FIELD = ("FIELD",)
    MATRIX = ("MATRIX",)
    POINT = ("POINT",)
    GEOMETRY = ("GEOMETRY",)
    RANGE = ("RANGE",)
    AUTH_CFG = "AUTH_CFG"
    SCALE = "SCALE"
    COLOR = "COLOR"
    LAYOUT = "LAYOUT"
    LAYOUT_ITEM = "LAYOUT_ITEM"
    DATETIME = "DATETIME"
    DATE = "DATE"
    TIME = "TIME"
    MAP_THEME = "MAP_THEME"
    PROVIDER_CONNECTION = "PROVIDER_CONNECTION"
    DATABASE_SCHEMA = "DATABASE_SCHEMA"
    DATABASE_TABLE = "DATABASE_TABLE"
    COORDINATE_OPERATION = "COORDINATE_OPERATION"
    POINTCLOUD_LAYER = "POINTCLOUD_LAYER"
    ANNOTATION_LAYER = "ANNOTATION_LAYER"

    def __init__(self):
        self._current = None
        self.instances = []

    def tr(self, string):
        """
        Returns a translatable string with the self.tr() function.
        """
        return QCoreApplication.translate("Processing", string)

    @property
    def current(self):
        return self._current

    @property
    def current_defined(self):
        return self._current is not None

    def __call__(self, *args, **kwargs):
        return self._define(*args, **kwargs)

    def _initnew(self):
        self._current = AlgWrapper()

    def _pop(self):
        self.instances.append(self.current)
        self._current = None

    def _define(self, *args, **kwargs):
        self._initnew()
        self.current.define(*args, **kwargs)

        def dec(f):
            self.current.set_func(f)
            self.current.end()
            self._pop()
            return f

        return dec

    def output(self, type, *args, **kwargs):
        """
        Define a output parameter for this algorithm using @alg.output.
        Apart from `type` this method will take all arguments and pass them though to the correct `QgsProcessingOutputDefinition` type.

        Types:

        - str: QgsProcessingOutputString
        - int: QgsProcessingOutputNumber
        - float: QgsProcessingOutputNumber
        - alg.NUMBER: QgsProcessingOutputNumber
        - alg.DISTANCE: QgsProcessingOutputNumber
        - alg.INT: QgsProcessingOutputNumber
        - alg.STRING: QgsProcessingOutputString
        - alg.FILE: QgsProcessingOutputFile
        - alg.FOLDER: QgsProcessingOutputFolder
        - alg.HTML: QgsProcessingOutputHtml
        - alg.LAYERDEF:  QgsProcessingOutputLayerDefinition
        - alg.MAPLAYER:  QgsProcessingOutputMapLayer
        - alg.MULTILAYER:  QgsProcessingOutputMultipleLayers
        - alg.RASTER_LAYER: QgsProcessingOutputRasterLayer
        - alg.VECTOR_LAYER: QgsProcessingOutputVectorLayer
        - alg.POINTCLOUD_LAYER: QgsProcessingOutputPointCloudLayer
        - alg.BOOL: QgsProcessingOutputBoolean

        :param type: The type of the input. This should be a type define on `alg` like alg.STRING, alg.DISTANCE
        :keyword label: The label of the output. Will convert into `description` arg.
        :keyword parent: The string ID of the parent parameter. Parent parameter must be defined before its here.
        """

        def dec(f):
            return f

        self.current.add_output(type, *args, **kwargs)
        return dec

    def help(self, helpstring, *args, **kwargs):
        """
        Define the help for the algorithm using @alg.help. This method will
        be used instead of the doc string on the function as the help in the processing dialogs.

        :param helpstring: The help text. Use QCoreApplication.translate() for translation support.
        """

        def dec(f):
            return f

        self.current.add_help(helpstring, *args, **kwargs)

        return dec

    def input(self, type, *args, **kwargs):
        """
        Define a input parameter for this algorithm using @alg.input.
        Apart from `type` this method will take all arguments and pass them though to the correct `QgsProcessingParameterDefinition` type.

        Types:

        - str: QgsProcessingParameterString
        - int: QgsProcessingParameterNumber
        - float: QgsProcessingParameterNumber
        - bool: QgsProcessingParameterBoolean
        - alg.NUMBER: QgsProcessingParameterNumber
        - alg.INT: QgsProcessingParameterNumber
        - alg.STRING: QgsProcessingParameterString
        - alg.DISTANCE: QgsProcessingParameterDistance
        - alg.SINK: QgsProcessingParameterFeatureSink
        - alg.SOURCE: QgsProcessingParameterFeatureSource
        - alg.FILE_DEST: QgsProcessingParameterFileDestination
        - alg.FOLDER_DEST: QgsProcessingParameterFolderDestination
        - alg.RASTER_LAYER: QgsProcessingParameterRasterLayer
        - alg.RASTER_LAYER_DEST: QgsProcessingParameterRasterDestination
        - alg.VECTOR_LAYER_DEST: QgsProcessingParameterVectorDestination
        - alg.POINTCLOUD_LAYER_DEST: QgsProcessingParameterPointCloudDestination
        - alg.BAND: QgsProcessingParameterBand
        - alg.BOOL: QgsProcessingParameterBoolean
        - alg.CRS: QgsProcessingParameterCrs
        - alg.ENUM: QgsProcessingParameterEnum
        - alg.EXPRESSION: QgsProcessingParameterExpression
        - alg.EXTENT: QgsProcessingParameterExtent
        - alg.FIELD: QgsProcessingParameterField
        - alg.FILE: QgsProcessingParameterFile
        - alg.MAPLAYER: QgsProcessingParameterMapLayer
        - alg.MATRIX: QgsProcessingParameterMatrix
        - alg.MULTILAYER: QgsProcessingParameterMultipleLayers
        - alg.POINT: QgsProcessingParameterPoint
        - alg.GEOMETRY: QgsProcessingParameterGeometry
        - alg.RANGE: QgsProcessingParameterRange
        - alg.VECTOR_LAYER: QgsProcessingParameterVectorLayer
        - alg.AUTH_CFG: QgsProcessingParameterAuthConfig
        - alg.MESH_LAYER: QgsProcessingParameterMeshLayer
        - alg.SCALE: QgsProcessingParameterScale
        - alg.LAYOUT: QgsProcessingParameterLayout
        - alg.LAYOUT_ITEM: QgsProcessingParameterLayoutItem
        - alg.COLOR: QgsProcessingParameterColor
        - alg.DATETIME: QgsProcessingParameterDateTime(type=QgsProcessingParameterDateTime.Type.DateTime)
        - alg.DATE: QgsProcessingParameterDateTime(type=QgsProcessingParameterDateTime.Type.Date)
        - alg.TIME: QgsProcessingParameterDateTime(type=QgsProcessingParameterDateTime.Type.Time)
        - alg.MAP_THEME: QgsProcessingParameterMapTheme
        - alg.PROVIDER_CONNECTION: QgsProcessingParameterProviderConnection
        - alg.DATABASE_SCHEMA: QgsProcessingParameterDatabaseSchema
        - alg.DATABASE_TABLE: QgsProcessingParameterDatabaseTable
        - alg.COORDINATE_OPERATION: QgsProcessingParameterCoordinateOperation
        - alg.POINTCLOUD_LAYER: QgsProcessingParameterPointCloudLayer
        - alg.ANNOTATION_LAYER: QgsProcessingParameterAnnotationLayer

        :param type: The type of the input. This should be a type define on `alg` like alg.STRING, alg.DISTANCE
        :keyword label: The label of the output. Translates into `description` arg.
        :keyword parent: The string ID of the parent parameter. Parent parameter must be defined before its here.
        :keyword default: The default value set for that parameter. Translates into `defaultValue` arg.
        """

        def dec(f):
            return f

        self.current.add_input(type, *args, **kwargs)

        return dec


input_type_mapping = {
    str: QgsProcessingParameterString,
    int: partial(
        QgsProcessingParameterNumber, type=QgsProcessingParameterNumber.Type.Integer
    ),
    float: partial(
        QgsProcessingParameterNumber, type=QgsProcessingParameterNumber.Type.Double
    ),
    bool: QgsProcessingParameterBoolean,
    ProcessingAlgFactory.NUMBER: partial(
        QgsProcessingParameterNumber, type=QgsProcessingParameterNumber.Type.Double
    ),
    ProcessingAlgFactory.INT: partial(
        QgsProcessingParameterNumber, type=QgsProcessingParameterNumber.Type.Integer
    ),
    ProcessingAlgFactory.STRING: QgsProcessingParameterString,
    ProcessingAlgFactory.DISTANCE: QgsProcessingParameterDistance,
    ProcessingAlgFactory.SINK: QgsProcessingParameterFeatureSink,
    ProcessingAlgFactory.SOURCE: QgsProcessingParameterFeatureSource,
    ProcessingAlgFactory.FILE_DEST: QgsProcessingParameterFileDestination,
    ProcessingAlgFactory.FOLDER_DEST: QgsProcessingParameterFolderDestination,
    ProcessingAlgFactory.RASTER_LAYER: QgsProcessingParameterRasterLayer,
    ProcessingAlgFactory.RASTER_LAYER_DEST: QgsProcessingParameterRasterDestination,
    ProcessingAlgFactory.VECTOR_LAYER_DEST: QgsProcessingParameterVectorDestination,
    ProcessingAlgFactory.POINTCLOUD_LAYER_DEST: QgsProcessingParameterPointCloudDestination,
    ProcessingAlgFactory.BAND: QgsProcessingParameterBand,
    ProcessingAlgFactory.BOOL: QgsProcessingParameterBoolean,
    ProcessingAlgFactory.CRS: QgsProcessingParameterCrs,
    ProcessingAlgFactory.ENUM: QgsProcessingParameterEnum,
    ProcessingAlgFactory.EXPRESSION: QgsProcessingParameterExpression,
    ProcessingAlgFactory.EXTENT: QgsProcessingParameterExtent,
    ProcessingAlgFactory.FIELD: QgsProcessingParameterField,
    ProcessingAlgFactory.FILE: QgsProcessingParameterFile,
    ProcessingAlgFactory.MAPLAYER: QgsProcessingParameterMapLayer,
    ProcessingAlgFactory.MATRIX: QgsProcessingParameterMatrix,
    ProcessingAlgFactory.MULTILAYER: QgsProcessingParameterMultipleLayers,
    ProcessingAlgFactory.POINT: QgsProcessingParameterPoint,
    ProcessingAlgFactory.GEOMETRY: QgsProcessingParameterGeometry,
    ProcessingAlgFactory.RANGE: QgsProcessingParameterRange,
    ProcessingAlgFactory.VECTOR_LAYER: QgsProcessingParameterVectorLayer,
    ProcessingAlgFactory.AUTH_CFG: QgsProcessingParameterAuthConfig,
    ProcessingAlgFactory.MESH_LAYER: QgsProcessingParameterMeshLayer,
    ProcessingAlgFactory.SCALE: QgsProcessingParameterScale,
    ProcessingAlgFactory.LAYOUT: QgsProcessingParameterLayout,
    ProcessingAlgFactory.LAYOUT_ITEM: QgsProcessingParameterLayoutItem,
    ProcessingAlgFactory.COLOR: QgsProcessingParameterColor,
    ProcessingAlgFactory.DATETIME: partial(
        QgsProcessingParameterDateTime,
        type=QgsProcessingParameterDateTime.Type.DateTime,
    ),
    ProcessingAlgFactory.DATE: partial(
        QgsProcessingParameterDateTime, type=QgsProcessingParameterDateTime.Type.Date
    ),
    ProcessingAlgFactory.TIME: partial(
        QgsProcessingParameterDateTime, type=QgsProcessingParameterDateTime.Type.Time
    ),
    ProcessingAlgFactory.MAP_THEME: QgsProcessingParameterMapTheme,
    ProcessingAlgFactory.PROVIDER_CONNECTION: QgsProcessingParameterProviderConnection,
    ProcessingAlgFactory.DATABASE_SCHEMA: QgsProcessingParameterDatabaseSchema,
    ProcessingAlgFactory.DATABASE_TABLE: QgsProcessingParameterDatabaseTable,
    ProcessingAlgFactory.COORDINATE_OPERATION: QgsProcessingParameterCoordinateOperation,
    ProcessingAlgFactory.POINTCLOUD_LAYER: QgsProcessingParameterPointCloudLayer,
    ProcessingAlgFactory.ANNOTATION_LAYER: QgsProcessingParameterAnnotationLayer,
}

output_type_mapping = {
    str: partial(_make_output, cls=QgsProcessingOutputString),
    int: partial(_make_output, cls=QgsProcessingOutputNumber),
    float: partial(_make_output, cls=QgsProcessingOutputNumber),
    ProcessingAlgFactory.NUMBER: partial(_make_output, cls=QgsProcessingOutputNumber),
    ProcessingAlgFactory.DISTANCE: partial(_make_output, cls=QgsProcessingOutputNumber),
    ProcessingAlgFactory.INT: partial(_make_output, cls=QgsProcessingOutputNumber),
    ProcessingAlgFactory.STRING: partial(_make_output, cls=QgsProcessingOutputString),
    ProcessingAlgFactory.FILE: partial(_make_output, cls=QgsProcessingOutputFile),
    ProcessingAlgFactory.FOLDER: partial(_make_output, cls=QgsProcessingOutputFolder),
    ProcessingAlgFactory.HTML: partial(_make_output, cls=QgsProcessingOutputHtml),
    ProcessingAlgFactory.LAYERDEF: partial(
        _make_output, cls=QgsProcessingOutputLayerDefinition
    ),
    ProcessingAlgFactory.MAPLAYER: partial(
        _make_output, cls=QgsProcessingOutputMapLayer
    ),
    ProcessingAlgFactory.MULTILAYER: partial(
        _make_output, cls=QgsProcessingOutputMultipleLayers
    ),
    ProcessingAlgFactory.RASTER_LAYER: partial(
        _make_output, cls=QgsProcessingOutputRasterLayer
    ),
    ProcessingAlgFactory.VECTOR_LAYER: partial(
        _make_output, cls=QgsProcessingOutputVectorLayer
    ),
    ProcessingAlgFactory.POINTCLOUD_LAYER: partial(
        _make_output, cls=QgsProcessingOutputPointCloudLayer
    ),
    ProcessingAlgFactory.BOOL: partial(_make_output, cls=QgsProcessingOutputBoolean),
}
